/****************************************************************/
/*          assignment4.c                                       */
/*          Written by Tzu-Kai Chou                             */
/****************************************************************/

#include <stdio.h>
#include <math.h>
#include <mpi.h>

/****************************************************************/
/*   Assume:                                                    */
/*   1) 2 sets of bodies of size N. (N=8)                       */
/*      X={1,2,3,4,5,6,7,8}                                     */
/*      Y={1,2,3,4,5,6,7,8}                                     */
/*   2) 4 processes. (NumProcess=4)                             */
/*   3) Each process should have n values. (n=N/NumProcess=2)   */
/****************************************************************/

int main(int argc, char *argv[])
{
    int NumProcess,MyRank;
    int i,j,k;
    
    int SumEach=0,SumTotal=0;
    int x[2],y[2],correlation[2];

    int dest,surc;
    MPI_Status status;

    int f(int a,int b); 

    /************************************************************/
    /*   Initialization                                         */
    /************************************************************/
    MPI_Init(&argc,&argv);
    MPI_Comm_size(MPI_COMM_WORLD,&NumProcess);
    MPI_Comm_rank(MPI_COMM_WORLD,&MyRank);

    if(NumProcess!=4)
    {
       printf("The program can only run on 4-node mode!");
       MPI_Finalize();
    }  
    else
    {
    /************************************************************/
    /*   Distribute x's & y's to processes                      */
    /************************************************************/
    switch(MyRank)
    {
     case 0:x[0]=1; x[1]=2;
            y[0]=1; y[1]=2;
            break; 
     case 1:x[0]=3; x[1]=4;
            y[0]=3; y[1]=4; 
            break;
     case 2:x[0]=5; x[1]=6;
            y[0]=5; y[1]=6; 
            break;
     case 3:x[0]=7; x[1]=8;
            y[0]=7; y[1]=8; 
            break;
    }

    /************************************************************/
    /*   Each process computes its own part.                    */
    /*   This is "Much Better Algorithm" described on           */
    /*   assignment 4 web page                                  */
    /************************************************************/
    for(k=0; k<NumProcess; k++)
    {
       /* Compute correlations of x's & y's in processor */
       for(j=0; j<2; j++)
          for(i=0; i<2; i++)
             correlation[i]+=f(x[i],y[j]);

       /* Communicate y's */
       /* Set up a ring of processes in one dimension. */
       /* P0 --> P1 --> P2 --> P3 -->P0                */
       dest=(MyRank+1)%NumProcess;
       surc=(MyRank-1+NumProcess)%NumProcess;
       MPI_Send(&y,2,MPI_INT,dest,0,MPI_COMM_WORLD);
       MPI_Recv(&y,2,MPI_INT,surc,0,MPI_COMM_WORLD,&status);
    }
    for(i=0; i<2; i++)
       SumEach+=correlation[i];  

    /************************************************************/
    /*   Process 0 gets total sum.                              */                    
    /************************************************************/
    MPI_Reduce(&SumEach,&SumTotal,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);

    /************************************************************/
    /*   Output the result                                      */
    /************************************************************/
    for(i=0; i<2; i++)
       printf("correlation[%d]=%d\n",i+MyRank*2,correlation[i]);
    if(MyRank==0)
       printf("The result=%d\n",SumTotal);

    /************************************************************/
    /*   Finalization                                           */
    /************************************************************/
    MPI_Finalize();
    }
}

/****************************************************************/
/*  Description : Hard-wired function                           */
/*                f(a,b)=abs( a*a + b*b )                       */
/*        Input : 2 integers a&b                                */
/*       Output : f(a,b)                                        */
/****************************************************************/
int f(int a,int b)
{
    return abs(a*a+b*b); 
}
